{
  "mappings": {
    "properties": {
      "{{ aggregate }}_id": {
        "type": "text",
        "fields": {
          "keyword": {
            "type": "keyword",
            "ignore_above": 256
          }
        }
      },
      "{{ aggregate }}_name": {
        "type": "text",
        "fields": {
          "keyword": {
            "type": "keyword",
            "ignore_above": 256
          }
        }
      },
      "{{ aggregate }}_country": {
        "type": "text",
        "fields": {
          "keyword": {
            "type": "keyword",
            "ignore_above": 256
          }
        }
      },
      "{{ aggregate }}_country_code": {
        "type": "text",
        "fields": {
          "keyword": {
            "type": "keyword",
            "ignore_above": 256
          }
        }
      },
      "{{ aggregate }}_region": {
        "type": "text",
        "fields": {
          "keyword": {
            "type": "keyword",
            "ignore_above": 256
          }
        }
      },
      "{{ aggregate }}_subregion": {
        "type": "text",
        "fields": {
          "keyword": {
            "type": "keyword",
            "ignore_above": 256
          }
        }
      },
      "{{ aggregate }}_coordinates": {
        "type": "geo_point"
      },
      "published_year": {
        "type": "date",
        "format": "yyyy-MM-dd"
      },
      "access_types_access_type": {
        "type": "text",
        "fields": {
          "keyword": {
            "type": "keyword",
            "ignore_above": 256
          }
        }
      },
      "access_types_status": {
        "type": "boolean"
      },
      "access_types_label": {
        "type": "text",
        "fields": {
          "keyword": {
            "type": "keyword",
            "ignore_above": 256
          }
        }
      },
      "access_types_total_outputs": {
        "type": "long"
      },
      "access_types_outputs_with_citations": {
        "type": "long"
      },
      "access_types_outputs_without_citations": {
        "type": "long"
      },
      "access_types_total_citations": {
        "type": "long"
      },
      "access_types_median_citations_per_output": {
        "type": "double"
      }
    }
  }
}