{# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Richard Hosking, James Diprose #}

SELECT 
  UPPER(TRIM(doi)) as doi,
  MAX( scopus_title ) as scopus_title,
  MAX( scopus_publicationName ) as scopus_publicationName,
  MAX( scopus_coverDate ) as scopus_coverDate,
  MAX( scopus_doi ) as scopus_doi,
  MAX( scopus_issn ) as scopus_issn,
  MAX( scopus_eIssn ) as scopus_eIssn,
  MAX( scopus_subtypeDescription ) as scopus_subtypeDescription,
  MAX( scopus_openaccess ) as scopus_openaccess,
  MAX( scopus_openaccessFlag ) as scopus_openaccessFlag,
  MAX( scopus_citedby_count ) as scopus_citedby_count,
  MAX( scopus_identifier ) as scopus_identifier,
  ARRAY_AGG(grid_id) as grids
FROM `{{ project_id }}.{{ dataset_id }}.scopus`
WHERE scopus IS NOT NULL AND doi IS NOT NULL
GROUP BY
  UPPER(TRIM(doi))