{# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Richard Hosking, James Diprose#}

SELECT
  doi,
  ARRAY(SELECT as STRUCT source, SUM(count) as count FROM UNNEST(months) GROUP BY source) as events,
  months,
  ARRAY(SELECT as STRUCT CAST(SPLIT(month, "-")[SAFE_OFFSET(0)] as int64) as year, source, SUM(count) as count FROM UNNEST(months) GROUP BY year, source) as years
FROM (
SELECT
  doi,
  ARRAY_AGG(STRUCT(month, source, count)) as months
FROM (
  SELECT
    (UPPER(TRIM(SUBSTR(obj_id, 17)))) as doi,
    safe.FORMAT_TIMESTAMP('%Y-%m', occurred_at) as month,
    source_id as source,
    COUNT(id) as count
  FROM `{{ project_id }}.{{ crossref_events.dataset_id }}.{{ crossref_events.table_id }}`
  WHERE safe.FORMAT_TIMESTAMP('%Y-%m', occurred_at) is not null
  GROUP BY
  doi, source_id, month)
GROUP BY doi)